/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.documents;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;

import org.htmlcleaner.CleanerProperties;
import org.htmlcleaner.HtmlCleaner;
import org.htmlcleaner.SimpleXmlSerializer;
import org.htmlcleaner.TagNode;
import org.htmlcleaner.XmlSerializer;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.tools.dom.IXHTMLNode;
import com.unitesk.requality.tools.dom.XHTMLResourceCollector;

class XMLNNode implements IXHTMLNode {
    private Node cur;

    public XMLNNode(Node root) {
        cur = root;
    }

    @Override
    public String getName() {
        return cur.getNodeName();
    }

    @Override
    public IXHTMLNode[] getChildren() {
        NodeList nl = cur.getChildNodes();
        IXHTMLNode[] result = new XMLNNode[nl.getLength()];
        for (int i = 0; i < nl.getLength(); i++) {
            result[i] = new XMLNNode(nl.item(i));
        }
        return result;
    }

    @Override
    public String[] getAttributeKeys() {
        NamedNodeMap nl = cur.getAttributes();
        String[] result = new String[nl.getLength()];
        for (int i = 0; i < nl.getLength(); i++) {
            result[i] = nl.item(i).getNodeName();
        }
        return result;
    }

    @Override
    public String getAttributeValue(String key) {
        if (cur.getAttributes().getNamedItem(key) == null) {
            return null;
        }
        return cur.getAttributes().getNamedItem(key).getNodeValue();
    }

    @Override
    public void setAttribute(String key, String value) {
        if (cur.getAttributes().getNamedItem(key) == null) {
            return;
        }
        cur.getAttributes().getNamedItem(key).setNodeValue(value);
    }
}

public class DocumentImporter {

    private IDocumentProcessor processor = null;

    public void setProcessor(IDocumentProcessor processor) {
        this.processor = processor;
    }

    public void importDocument(String docId, File inFile, File outFile,
            TreeDB base, TreeNode node, String encoding) throws Exception {
        Reader reader = new InputStreamReader(new FileInputStream(inFile),
                encoding);

        if (!outFile.getParentFile().exists())
            outFile.getParentFile().mkdirs();

        Writer writer = new OutputStreamWriter(new FileOutputStream(outFile),
                "UTF-8");

        try {
            StringWriter out = new StringWriter();
            cleanHtml(reader, out);
            out.close();
            reader.close();

            if (processor != null) {
                processor.process(docId, node);
            } else {
                writer.append(out.toString());
            }
            writer.close();
            final Document doc = (Document) base.getNode(docId);
            doc.setMetaCharset("UTF-8");
            IXHTMLNode rootNode = new XMLNNode(doc.getDOM().getFirstChild());
            String srcHtmlFolder = inFile.getParent();
            String nodeResourceFolder = new File(doc.getHTMLPath()).getParent();
            XHTMLResourceCollector docProc = new XHTMLResourceCollector(
                    rootNode, srcHtmlFolder, nodeResourceFolder);
            docProc.process();
            doc.saveDOM();
            /*
             * NodeEditorInput input = new NodeEditorInput(doc); try {
             * IDE.openEditor
             * (PlatformUI.getWorkbench().getActiveWorkbenchWindow()
             * .getActivePage(), input, ReqMarker.ID); } catch
             * (PartInitException e) { e.printStackTrace(); }
             */
        } catch (Exception e) {
            e.printStackTrace();
            throw e;
        } finally {
            reader.close();
            writer.close();
        }
    }

    public static void cleanHtml(Reader intHtml, Writer out) throws IOException {
        CleanerProperties props = createCleanerProperties();
        HtmlCleaner cleaner = new HtmlCleaner(props);
        TagNode root = cleaner.clean(intHtml);
        root.addAttribute("xmlns", "http://www.w3.org/1999/xhtml");
        if (root.hasAttribute("xmlns:xml")) {
            root.removeAttribute("xmlns:xml");
        }
        XmlSerializer serializer = new SimpleXmlSerializer(props);
        root.serialize(serializer, out);
    }

    private static CleanerProperties createCleanerProperties() {
        CleanerProperties properties = new CleanerProperties();
        properties.setAdvancedXmlEscape(true);
        properties.setTranslateSpecialEntities(true);
        properties.setRecognizeUnicodeChars(true);
        properties.setUseEmptyElementTags(false);
        properties.setNamespacesAware(true);
        properties.setUseCdataForScriptAndStyle(true);
        properties.setOmitDoctypeDeclaration(false);
        properties.setIgnoreQuestAndExclam(true);
        return properties;
    }
}