/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.core;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.tools.TreesTracker;

public class DeltaProcessor implements IResourceChangeListener {
    private Map<IFile, UUID> resourceToNode;
    private ResourceStorage rs;
    private Map<UUID, IFile> added = new LinkedHashMap<UUID, IFile>();
    private Map<UUID, IFile> deleted = new LinkedHashMap<UUID, IFile>();
    private Map<UUID, IFile> changed = new LinkedHashMap<UUID, IFile>();
    private Map<UUID, IFile[]> moved = new LinkedHashMap<UUID, IFile[]>();

    public DeltaProcessor(Map<IFile, UUID> resourceToNode, ResourceStorage rs) {
        this.resourceToNode = resourceToNode;
        this.rs = rs;
    }

    @Override
    public void resourceChanged(IResourceChangeEvent event) {
        if (event.getType() == IResourceChangeEvent.PRE_DELETE
                || event.getType() == IResourceChangeEvent.PRE_CLOSE) {
            return;
        }
        if (event.getDelta().getResource().getType() == IResource.ROOT) {
            for (IResourceDelta rd : event.getDelta().getAffectedChildren()) {
                if (rd.getResource().getProject().equals(rs.getProject())) {
                    processDelta(rd);
                }
            }
        } else {
            throw new IllegalArgumentException(
                    "Delta resource must be root type");
        }
    }

    private void processChange(IFile file, int changeType) {
        UUID uuid;
        switch (changeType) {
        case IResourceDelta.ADDED:
            uuid = rs.getUUId(file);
            added.put(uuid, file);
            break;
        case IResourceDelta.REMOVED:
            uuid = resourceToNode.get(file);
            deleted.put(uuid, file);
            break;
        case IResourceDelta.CHANGED:
            uuid = rs.getUUId(file);
            changed.put(uuid, file);
            break;
        default:
            throw new RuntimeException();
        }
    }

    public void processDelta(IResourceDelta delta) {
        moved.clear();
        deleted.clear();
        added.clear();
        changed.clear();

        Map<IFile, Integer> result = collectChanges(delta, null);
        for (IFile f : result.keySet()) {
            processChange(f, result.get(f));
        }

        findNodeMove();

        for (UUID uuid : changed.keySet()) {
            if (changed.get(uuid).exists()) {
                rs.getDeltaTarget().changed(uuid);
            }
        }
        // apply changes to cache
        for (UUID uuid : deleted.keySet()) {
            rs.getDeltaTarget().deleted(uuid);
            resourceToNode.remove(deleted.get(uuid));
        }

        for (UUID uuid : added.keySet()) {
            resourceToNode.put(added.get(uuid), uuid);
            rs.getDeltaTarget().created(uuid);
        }

        for (UUID uuid : moved.keySet()) {
            IFile to = moved.get(uuid)[0];
            IFile from = moved.get(uuid)[1];
            if (from.getParent().equals(to.getParent())) { // rename move
                resourceToNode.remove(from);
                resourceToNode.put(to, uuid);
                rs.getDeltaTarget().changed(uuid);
            } else {
                String opId = from.getParent().getName();
                IFile oldParentFile = from.getParent().getParent()
                        .getFile(new Path(opId + ".json"));
                UUID fromParent = getUUIdFromMoved(oldParentFile);
                if (fromParent == null) {
                    fromParent = resourceToNode.get(oldParentFile);
                }
                resourceToNode.put(to, uuid);
                resourceToNode.remove(from);
                rs.getDeltaTarget().moved(uuid, fromParent, rs.getParent(to));
            }
        }
    }

    private void findNodeMove() {
        for (UUID uuid : deleted.keySet()) {
            if (added.get(uuid) != null) {
                moved.put(uuid,
                        new IFile[] { added.get(uuid), deleted.get(uuid) });
            }
        }

        for (UUID uuid : moved.keySet()) {
            added.remove(uuid);
            deleted.remove(uuid);
        }

        Map<UUID, IFile[]> result = new HashMap<UUID, IFile[]>();

        for (UUID uuid : moved.keySet()) {
            IFile[] f = moved.get(uuid);
            if (!(isAbsorbed(f[0]) || isAbsorbed(f[1]))) {
                result.put(uuid, f);
            }
            resourceToNode.put(f[0], uuid);
            resourceToNode.remove(f[1]);
        }
        moved = result;
    }

    private boolean isAbsorbed(IFile f) {
        for (UUID uuid : moved.keySet()) {
            IFile chFiles[] = moved.get(uuid);
            IPath f1p = chFiles[0]
                    .getParent()
                    .getFullPath()
                    .addTrailingSeparator()
                    .append(chFiles[0].getName().substring(0,
                            chFiles[1].getName().length() - 5));
            IPath f2p = chFiles[1]
                    .getParent()
                    .getFullPath()
                    .addTrailingSeparator()
                    .append(chFiles[1].getName().substring(0,
                            chFiles[1].getName().length() - 5));
            ;
            if (f1p.isPrefixOf(f.getFullPath())) {
                if (!f.equals(chFiles[0])) {
                    return true;
                }
            }
            if (f2p.isPrefixOf(f.getFullPath())) {
                if (!f.equals(chFiles[1])) {
                    return true;
                }
                return true;
            }
        }
        return false;
    }

    public UUID getUUIdFromMoved(IFile file) {
        for (UUID uuid : moved.keySet()) {
            if (moved.get(uuid)[0].equals(file)) {
                return uuid;
            }
            if (moved.get(uuid)[1].equals(file)) {
                return uuid;
            }
        }
        return null;
    }

    private boolean isNode(IResource resource) {
        if (resourceToNode.containsKey(resource)) {
            return true;
        }
        if (resource.getType() == IResource.FILE) {
            IFile file = (IFile) resource;
            if (file.getFileExtension() != null
                    && file.getFileExtension().equals("json")) {
                return true;
            }
        }
        return false;
    }

    private Map<IFile, Integer> collectChanges(IResourceDelta delta,
            Map<IFile, Integer> changes) {
        Map<IFile, Integer> result = changes;
        if (result == null) {
            result = new LinkedHashMap<IFile, Integer>();
        }

        IResourceDelta subDeltas[] = delta.getAffectedChildren();

        for (IResourceDelta sub : subDeltas) {
            IResource res = sub.getResource();
            if ((sub.getKind() & IResourceDelta.ADDED) != 0) {
                TreeNode node;
                if (isNode(res)) {
                    result.put((IFile) res, IResourceDelta.ADDED);
                } else if ((node = getAffectedNode(res)) != null) {
                    IFile file = (IFile) TreesTracker.getResourceByNode(node);
                    result.put(file, IResourceDelta.CHANGED);
                }
            }

            if ((sub.getKind() & IResourceDelta.CHANGED) != 0) {
                if (sub.getAffectedChildren().length == 0) {
                    TreeNode node = getAffectedNode(res);
                    if (node != null) {
                        IResource dres = TreesTracker.getResourceByNode(node);
                        result.put((IFile) dres, IResourceDelta.CHANGED);
                    }
                }
            }

            collectChanges(sub, result);

            if ((sub.getKind() & IResourceDelta.REMOVED) != 0) {
                TreeNode node;
                if ((node = getAffectedNode(res)) != null) {
                    IFile file = (IFile) TreesTracker.getResourceByNode(node);
                    if(result.get(file)==null) {
                        result.put(file, IResourceDelta.CHANGED);
                    }
                }
                if (isNode(res)) {
                    result.put((IFile) res, IResourceDelta.REMOVED);
                }
            }
        }
        return result;
    }

    private TreeNode getAffectedNode(IResource res) {
        if (isNode(res)) {
            return TreesTracker.getNodeByResource((IFile) res);
        }
        TreeDB db = TreesTracker.getTracker().getTree(res.getProject());
        if (db.getRootNode(false) == null) {
            return null; // new project
        }
        IResource root = TreesTracker.getResourceByNode(db.getRootNode(false));
        IPath path = res.getLocation().removeFirstSegments(
                root.getParent().getLocation().segmentCount() + 1);
        TreeNode node = db.getRootNode(false);
        for (String segment : path.segments()) {
            if (segment.endsWith("_resources")) {
                String nodeName = segment.substring(0, segment.length()
                        - "_resources".length());
                if (node.findChild(nodeName) != null) {
                    return node.findChild(nodeName);
                }
            }
            node = node.findChild(segment);
            if (node == null) {
                return null;
            }
        }
        return node;
    }
}
