/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.ui.cnf;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.navigator.CommonNavigator;
import org.eclipse.ui.navigator.CommonViewer;
import org.eclipse.ui.views.properties.IPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.ITabbedPropertySheetPageContributor;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;

import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.ResourceStorage;
import com.unitesk.requality.eclipse.tools.ITreeStateListener;
import com.unitesk.requality.eclipse.tools.IUIStateChangeListener;
import com.unitesk.requality.eclipse.tools.TreesTracker;

public class RequalityCNF extends CommonNavigator implements
        ITabbedPropertySheetPageContributor {
    public final static String ID = "com.unitesk.requality.views.RequalityExplorer";

    ITreeStateListener treeStateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            db.addListener(treeChangeListener);
        }
        
        @Override
        public void delete(IProject proj) {
        }
        
        @Override
        public void close(TreeDB db) {
            db.removeListener(treeChangeListener);
        }
    };
    ITreeChangeListener treeChangeListener = new ITreeChangeListener() {
        @Override
        public void moved(TreeNode node, TreeNode oldParent) {
            refresh(oldParent);
            refresh(node.getParent());
        }

        @Override
        public void deleted(TreeNode node) {
            refresh(node.getParent());
        }

        @Override
        public void changed(TreeNode node) {
            refresh(node.getParent());
        }

        @Override
        public void added(final TreeNode node) {
            if (node.getParent() == null) {
                refresh(null);
            } else {
                refresh(node.getParent());
            }
            // TODO: smart tree expand 
            getCommonViewer().getControl().getDisplay().asyncExec(new Runnable() {
                @Override
                public void run() {
                    if (getCommonViewer().getControl().isDisposed()) {
                        return;
                    }
                    getCommonViewer().expandToLevel(node, 1);
                }
            });

        }
    };

    final private IUIStateChangeListener uiListener = new IUIStateChangeListener() {
        @Override
        public void newActiveNode(TreeNode node) {
            if (node == null) {
                return;
            }
            Object obj = ((IStructuredSelection) getCommonViewer()
                    .getSelection()).getFirstElement();
            if (obj != null && obj.equals(node))
                return;
            Object parentElement = node.getParent();
            StructuredSelection element = new StructuredSelection(node);
            getCommonViewer().refresh(node.getParent());
            getCommonViewer().expandToLevel(parentElement, 1);
            selectReveal(element);
        }
    };

    final private ISelectionChangedListener treeSelection = new ISelectionChangedListener() {
        @Override
        public void selectionChanged(SelectionChangedEvent event) {
            if (event.getSelection() instanceof IStructuredSelection) {
                Object obj = ((IStructuredSelection) event.getSelection())
                        .getFirstElement();
                if (obj instanceof TreeNode) {
                    TreeNode node = (TreeNode) obj;
                    if (node == null
                            || node.equals(Activator.getDefault()
                                    .getActiveNode(node.getType())))
                        return;
                    Activator.getDefault().setActiveNode(node, node.getType());
                }
            }
        }
    };

    public RequalityCNF() {
        super();
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.addListener(treeChangeListener);
        }
        RequalityPlugin.getTreesTracker().addListener(treeStateListener);
    }

    @Override
    public void dispose() {
        super.dispose();
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.removeListener(treeChangeListener);
        }
        getCommonViewer().removePostSelectionChangedListener(treeSelection);
        Activator.getDefault().removeChangeListener(uiListener);
        RequalityPlugin.getTreesTracker().removeListener(treeStateListener);
    }

    public synchronized void refresh(final TreeNode treeNode) {
        getCommonViewer().getControl().getDisplay().asyncExec(new Runnable() {
            @Override
            public void run() {
                if (getCommonViewer().getControl().isDisposed()) {
                    return;
                }
                if (treeNode == null || (treeNode instanceof IProject)
                        || treeNode.getTreeDB().getRootNode() == treeNode) {
                    getCommonViewer().refresh();
                    return;
                }
                getCommonViewer().refresh(true);
                getCommonViewer().refresh(treeNode);
            }
        });
    }

    @Override
    public void createPartControl(Composite aParent) {
        super.createPartControl(aParent);
        Activator.getDefault().addChangeListener(uiListener);
        getCommonViewer().addPostSelectionChangedListener(treeSelection);
        getCommonViewer().setSorter(new ViewerSorter() {
            @Override
            public int compare(Viewer viewer, Object e1, Object e2) {
                if (!(e1 instanceof TreeNode)) {
                    return super.compare(viewer, e1, e2);
                }
                if (!(e2 instanceof TreeNode)) {
                    return super.compare(viewer, e1, e2);
                }
                TreeNode n1 = (TreeNode) e1;
                TreeNode n2 = (TreeNode) e2;
                if (n1.getType().compareTo(n2.getType()) != 0) {
                    return n1.getType().compareTo(n2.getType());
                }
                return n1.getId().compareTo(n2.getId());
            }
        });
    }

    // This is danger hack
    @Override
    protected CommonViewer createCommonViewerObject(Composite aParent) {
        CommonViewer cv = new CommonViewer(getViewSite().getId(), aParent,
                SWT.H_SCROLL | SWT.V_SCROLL);
        return cv;
    }

    @Override
    public String getContributorId() {
        return "com.unitesk.requality.views.RequalityPropertyContributor";
    }

    @Override
    public Object getAdapter(@SuppressWarnings("rawtypes") Class adapter) {
        if (adapter == IPropertySheetPage.class) {
            return new TabbedPropertySheetPage(this);
        }
        return super.getAdapter(adapter);
    }

    public static synchronized TreeNode getSelectedNode() {

        IWorkbenchWindow activeWorkbenchWindow = PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow();
        if (activeWorkbenchWindow == null)
            return null;
        ISelection selection = activeWorkbenchWindow.getSelectionService()
                .getSelection();

        if (!(selection instanceof ITreeSelection)) {
            return null;
        }
        ITreeSelection selected = (ITreeSelection) selection;

        if (selected.isEmpty()) {
            return null;
        }

        Object obj = selected.getFirstElement();
        if (obj instanceof TreeNode) {
            TreeNode node = (TreeNode) obj;
            return node;
        } else if (obj instanceof IFile) {
            IFile f = (IFile) obj;
            TreeDB db = TreesTracker.getTracker().getTree(f.getProject());
            ResourceStorage rs = (ResourceStorage) db.getStorage();
            if (rs.getUUId(f) == null) {
                return null;
            }
            return db.getNode(rs.getUUId(f));
        }
        return null;
    }

    @Override
    protected Object getInitialInput() {
        if(super.getInitialInput()==null) {
            return ResourcesPlugin.getWorkspace().getRoot();
        }
        return super.getInitialInput();
    }
}
