/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.documents;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.ui.dialogs.SelectDocDialog;
import com.unitesk.requality.merge.MergeProcessor;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public class LocationTracker extends DocumentProcessor {
    private Document oldDocument = null;
    private TreeDB db;
    private String[] mergedLocations;

    @Override
    public boolean configure(final Document doc) {
        this.db = doc.getTreeDB();
        Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
                .getShell();
        SelectDocDialog dialog = new SelectDocDialog(shell, db) {
            @Override
            public boolean canSelect() {
                if(!super.canSelect()) return false;
                if(!selectedNode.getType().equals("Document")) return false;
                if(selectedNode.getQualifiedId().equals(doc.getQualifiedId())) return false;
                return true;
            }
        };
        if (dialog.open() == SelectDocDialog.OK) {
            oldDocument = (Document) dialog.getResult();
        }
        return false;
    }

    @Override
    protected void process(String docId, final Reader reader,
            final Writer writer, TreeNode rootNode) throws IOException {
        Document newDoc = (Document) db.getNode(docId);
        final InputStreamReader oldIn = new InputStreamReader(
                new FileInputStream(oldDocument.getHTMLPath()), "UTF-8");
        final MergeProcessor mp = MergeProcessor.getInstance();
        try {
            Shell shell = PlatformUI.getWorkbench().getActiveWorkbenchWindow()
                    .getShell();
            ProgressMonitorDialog dialog = new ProgressMonitorDialog(shell);
            dialog.run(true, false, new IRunnableWithProgress() {
                public void run(IProgressMonitor monitor) {
                    monitor.beginTask("Updating document...",
                            IProgressMonitor.UNKNOWN);
                    mergedLocations = mp.merge(oldIn, reader, writer);
                    monitor.done();
                }
            });
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            MessageDialog.openInformation(PlatformUI.getWorkbench()
                    .getActiveWorkbenchWindow().getShell(),
                    "Document processing report", "Internal error occured");
        } catch (InterruptedException e) {
            e.printStackTrace();
            MessageDialog.openInformation(PlatformUI.getWorkbench()
                    .getActiveWorkbenchWindow().getShell(),
                    "Document processing report", "Internal error occured");
        }

        oldIn.close();
        int processedLocations = 0;
        for (String id : mergedLocations) {
            Location oldLoc = oldDocument.getLocation(id);
            if (oldLoc != null) {
                Requirement[] reqs = oldLoc.getRequirements();
                Location loc = (Location) newDoc.createChild(id, "Location");
                loc.setDescription(oldLoc.getDescription());
                loc.saveAttributes();
                for (Requirement req : reqs) {
                    req.addLocation(loc.getQualifiedId());
                    req.saveAttributes();
                }
                processedLocations++;
            }
        }
        MessageDialog.openInformation(PlatformUI.getWorkbench()
                .getActiveWorkbenchWindow().getShell(),
                "Document processing report", processedLocations
                        + " locations processed");
    }
}
