/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FormAttachment;
import org.eclipse.swt.layout.FormData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.ui.dialogs.EditAttrDialog;
import com.unitesk.requality.nodetypes.Requirement;

public class AttributesPanel {
    private final class AttributeTableContentProvider implements
            IStructuredContentProvider {
        @Override
        public Object[] getElements(Object input) {
            if (input instanceof TreeNode) {
                TreeNode node = (TreeNode) input;
                ArrayList<Attribute> data = new ArrayList<Attribute>();
                for (String key : node.getAttributeKeys()) {
                    if (!key.startsWith("_")) {
                        data.add(node.getAttribute(key));
                    }
                }
                return data.toArray(new Object[0]);
            }
            return null;
        }

        @Override
        public void dispose() {
        }

        @Override
        public void inputChanged(Viewer arg0, Object arg1, Object arg2) {
        }
    }

    private Composite panel;
    private static final String columnNames[] = new String[] { "Name", "Value" };
    private Requirement req;
    private TableViewer attribTable;
    private Button addAttrButton;
    private Button removeAttrButton;

    private Listener addAttrListener = new Listener() {
        public void handleEvent(Event event) {
            if ((event.type == SWT.KeyDown && event.keyCode == SWT.INSERT)
                    || (event.type == SWT.Selection)) {

                EditAttrDialog addAttrDialog = new EditAttrDialog(
                        panel.getShell(), "", "", makeAttrMap());
                if (addAttrDialog.open() == EditAttrDialog.OK) {
                    req.putAttribute(new Attribute(AttributeType.STRING,
                            addAttrDialog.getKey(), addAttrDialog.getValue()));
                    attribTable.refresh();
                    attribTable.getControl().setFocus();
                    panel.notifyListeners(SWT.Modify, new Event());
                    if (req != null) {
                        req.putAttribute(new Attribute(AttributeType.STRING,
                                addAttrDialog.getKey(), addAttrDialog
                                        .getValue()));
                        req.saveAttributes();
                    }
                }
            }
        }
    };

    private Listener removeAttrLink = new Listener() {
        @Override
        public void handleEvent(Event event) {
            if ((event.type == SWT.KeyDown && event.keyCode == SWT.DEL)
                    || (event.type == SWT.Selection)) {

                TableItem[] items = attribTable.getTable().getSelection();
                String mes = items.length <= 1 ? "Are you sure you want to remove selected attribute?"
                        : "Are you sure you want to remove selected attributes?";

                if ((items.length > 0)
                        && (MessageDialog.openQuestion(panel.getShell(),
                                "Attribute Editor", mes) == true)) {

                    for (int i = 0; i < items.length; i++) {
                        Object attrName = ((Attribute) items[i].getData()).getKey();
                        if (req != null) {
                            req.removeAttribute((String) attrName);
                            req.saveAttributes();
                        }
                        attribTable.refresh();
                        attribTable.getControl().setFocus();
                        panel.notifyListeners(SWT.Modify, new Event());
                    }
                }
            }
        }
    };

    private Map<String, Object> makeAttrMap() {
        Map<String, Object> innerAttr = new HashMap<String, Object>();
        for (String key : req.getAttributeKeys()) {
            innerAttr.put(key, req.getAttributeValue(key).toString());
        }
        return innerAttr;
    }

    // TODO: refactor magic numbers
    public AttributesPanel(Composite parent, TabbedPropertySheetWidgetFactory wf) {
        panel = wf.createComposite(parent, SWT.NONE);
        panel.setLayout(new FormLayout());

        FormData labelData = new FormData();
        labelData.top = new FormAttachment(0, 4);
        Label label = wf.createLabel(panel, "Attributes:", SWT.NONE);
        label.setLayoutData(labelData);

        FormData buttonsData = new FormData();
        buttonsData.left = new FormAttachment(label, 10);
        Composite attrsButtons = wf.createComposite(panel, SWT.NONE);
        buttonsData.height = 30;
        attrsButtons.setLayoutData(buttonsData);
        attrsButtons.setLayout(new GridLayout(2, false));

        GridData compactButton = new GridData();
        compactButton.heightHint = 28;
        compactButton.verticalIndent = -4;
        addAttrButton = wf.createButton(attrsButtons, "Add...", SWT.NONE);
        addAttrButton.setToolTipText("Add attribute");
        addAttrButton.addListener(SWT.Selection, addAttrListener);
        addAttrButton.setLayoutData(compactButton);

        removeAttrButton = wf.createButton(attrsButtons, "Remove...", SWT.NONE);
        removeAttrButton.setToolTipText("Remove attribute");
        removeAttrButton.addListener(SWT.Selection, removeAttrLink);
        removeAttrButton.setLayoutData(compactButton);
        removeAttrButton.setEnabled(false);

        attribTable = new TableViewer(panel, SWT.H_SCROLL | SWT.V_SCROLL
                | SWT.BORDER | SWT.RESIZE);
        attribTable.setSorter(new ViewerSorter() {
            private int getVal(String s) {
                int a = 0;
                for (int i = 0; i < s.length() && s.charAt(i) >= '0'
                        && s.charAt(i) <= '9'; i++)
                    a = a * 10 + s.charAt(i) - '0';
                return a;
            }

            public int compare(Viewer viewer, Object e1, Object e2) {
                if (e1 instanceof Object[] && e2 instanceof Object[]) {
                    Object[] ar1 = (Object[]) e1, ar2 = (Object[]) e2;
                    if (ar1.length != ar2.length)
                        return ar1.length - ar2.length;
                    for (int i = 0; i < ar1.length; i++) {
                        String s1 = ar1[i].toString(), s2 = ar2[i].toString();
                        if (s1 != null && s1.length() > 0 && s2 != null
                                && s2.length() > 0) {
                            int a1 = getVal(s1), a2 = getVal(s2);
                            if (a1 != a2 && s1.charAt(0) >= '0'
                                    && s1.charAt(0) <= '9'
                                    && s2.charAt(0) >= '0'
                                    && s2.charAt(0) <= '9')
                                return a1 - a2;
                        }
                        @SuppressWarnings("unchecked")
                        int res = getComparator().compare(s1, s2);
                        if (res != 0)
                            return res;
                    }
                    return 0;
                }
                return super.compare(viewer, e1, e2);
            }
        });
        FormData tableData = new FormData(0, 0);
        tableData.top = new FormAttachment(0, 30);
        tableData.left = new FormAttachment(0, 5);
        tableData.bottom = new FormAttachment(100, -5);
        tableData.right = new FormAttachment(100, -5);
        attribTable.getControl().setLayoutData(tableData);

        // Create columns
        for (String colName : columnNames) {
            TableViewerColumn column = new TableViewerColumn(attribTable,
                    SWT.H_SCROLL | SWT.V_SCROLL);
            column.getColumn().setText(colName);
            column.getColumn().setWidth(100);
            column.getColumn().setResizable(true);
            column.getColumn().setMoveable(true);
        }
        attribTable.setColumnProperties(columnNames);
        attribTable.getTable().setHeaderVisible(true);
        attribTable.getTable().setLinesVisible(true);
        // Editors
        CellEditor[] editors = new CellEditor[columnNames.length];
        editors[0] = new TextCellEditor(attribTable.getTable());
        editors[1] = new TextCellEditor(attribTable.getTable());
        attribTable.setCellEditors(editors);

        attribTable.setLabelProvider(new ITableLabelProvider() {
            @Override
            public void removeListener(ILabelProviderListener arg0) {
            }

            @Override
            public void addListener(ILabelProviderListener arg0) {
            }

            @Override
            public boolean isLabelProperty(Object arg0, String arg1) {
                return true;
            }

            @Override
            public void dispose() {
            }

            @Override
            public String getColumnText(Object element, int columnIndex) {
                Attribute attr = (Attribute) element;
                if (columnIndex == 0) {
                    return attr.getKey();
                }
                if (columnIndex == 1) {
                    return attr.getValue().toString();
                }
                return "n/a";
            }

            @Override
            public Image getColumnImage(Object element, int columnIndex) {
                return null;
            }
        });
        attribTable.setContentProvider(new AttributeTableContentProvider());

        attribTable.addDoubleClickListener(new IDoubleClickListener() {
            @Override
            public void doubleClick(DoubleClickEvent event) {
                Attribute attribute = (Attribute) ((IStructuredSelection) event
                        .getSelection()).getFirstElement();
                String savedOldName = (String) attribute.getKey();
                EditAttrDialog addAttrDialog = new EditAttrDialog(getControl()
                        .getShell(), (String) attribute.getKey(),
                        (String) attribute.getValue(), makeAttrMap());
                if (addAttrDialog.open() == EditAttrDialog.OK) {
                    if (!savedOldName.equals(addAttrDialog.getKey())) {
                        req.removeAttribute(savedOldName);
                    }
                    req.putAttribute(new Attribute(AttributeType.STRING,
                            addAttrDialog.getKey(), addAttrDialog.getValue()));
                    req.saveAttributes();
                    attribTable.refresh();
                    attribTable.getControl().setFocus();
                    panel.notifyListeners(SWT.Modify, new Event());
                    attribTable.refresh();
                }
            }
        });
        attribTable
                .addSelectionChangedListener(new ISelectionChangedListener() {
                    @Override
                    public void selectionChanged(SelectionChangedEvent event) {
                        removeAttrButton.setEnabled(!attribTable.getSelection()
                                .isEmpty());
                    }
                });
    }

    public Control getControl() {
        return panel;
    }

    public void setRequirement(Requirement req) {
        this.req = req;
        attribTable.setInput(req);
    }
}
