/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import org.eclipse.core.resources.IProject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.views.properties.tabbed.AbstractPropertySection;

import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.tools.ITreeStateListener;
import com.unitesk.requality.eclipse.tools.TreesTracker;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public abstract class TreeNodePropertiesTab extends AbstractPropertySection {
    protected TreeNode node;
    boolean isDisposed = false;

    ITreeChangeListener treeChangeListener = new ITreeChangeListener() {
        @Override
        public void moved(TreeNode node, TreeNode oldParent) {
            if (node != null)
                setTreeNodeSafe(node);
        }

        @Override
        public void deleted(TreeNode node) {
            if (node != null)
                setTreeNodeSafe(node);
        }

        @Override
        public void changed(TreeNode node) {
            if (node != null)
                setTreeNodeSafe(node);
        }

        @Override
        public void added(TreeNode node) {
            if (node != null)
                setTreeNodeSafe(node);
        }
    };

    ITreeStateListener treeStateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            db.addListener(treeChangeListener);
            if (node != null)
                setTreeNodeSafe(node);
        }

        @Override
        public void close(TreeDB db) {
            db.removeListener(treeChangeListener);
            if (node != null)
                setTreeNodeSafe(node);
        }

        @Override
        public void delete(IProject proj) {
            // TODO Auto-generated method stub
            
        }
    };

    public TreeNodePropertiesTab() {
        super();
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.addListener(treeChangeListener);
        }
        TreesTracker.getTracker().addListener(treeStateListener);
    }

    @Override
    public void dispose() {
        isDisposed = true;
        for (TreeDB db : TreesTracker.getTracker().getTrees()) {
            db.removeListener(treeChangeListener);
        }
        TreesTracker.getTracker().removeListener(treeStateListener);
        super.dispose();
    }

    private void setTreeNodeSafe(final TreeNode node) {
        Display d = Display.getCurrent();
        if (d == null) {
            d = Display.getDefault();
        }
        d.syncExec(new Runnable() {
            @Override
            public void run() {
                if (!isDisposed) {
                    setTreeNode(node);
                }
            }
        });
    }

    abstract protected void setTreeNode(TreeNode node);

    @Override
    public void setInput(IWorkbenchPart part, ISelection selection) {
        super.setInput(part, selection);
        node = TreeNodePropertiesTab.toTreeNode(selection);
        if (node != null) {
            setTreeNodeSafe(node);
        } else {
            // Unsupported input
        }
    }

    protected Requirement toRequirement(TreeNode node) {
        if (node instanceof Requirement) {
            return (Requirement) node;
        }

        if (node instanceof Location) {
            Location location = (Location) node;
            Requirement[] reqs = location.getRequirements();
            if (reqs.length > 0) {
                return reqs[0];
            }
        }
        return null;
    }

    public static TreeNode toTreeNode(ISelection sel) {
        TreeNode node = (TreeNode) ((IStructuredSelection) sel)
                .getFirstElement();
        if (node instanceof TreeNode) {
            return (TreeNode) node;
        }
        return null;
    }

}
