/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.UUID;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;

public class Location extends TreeNode {
    private String label;

	public Location() {
        super(null, null);
    }

    public Location(TreeDB db, UUID uid) {
        super(db, uid);
        this.id = uid.toString();
        setType("Location");
    }

    public String getDescription(String basePath) {
        String result = getDescription("<div>", "</div>", basePath);
        return result;
    }

    public Requirement[] getRequirements() {
        String locQId = getQualifiedId();
        return getTreeDB().getRequirementsFor(locQId);
    }

    public static String getDescription(String htmlFolderPath,
            String description, String prefix, String postfix, String basePath) {
        ByteArrayOutputStream result = new ByteArrayOutputStream();
        try {
            DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance()
                    .newDocumentBuilder();
            String content = prefix + description + postfix;

            org.w3c.dom.Document doc = docBuilder
                    .parse(new ByteArrayInputStream(content.getBytes("UTF-8")));
            String relPath = calcRelPath(htmlFolderPath, basePath) + "/";
            walk(doc.getFirstChild(), relPath, htmlFolderPath);
            DocumentFragment fragment = doc.createDocumentFragment();

            NodeList childNodes = doc.getFirstChild().getChildNodes();
            int len = childNodes.getLength();
            for (int i = 0; i < len; i++) {
                // appended child is removed from original location
                // so we always add the first child
                fragment.appendChild(childNodes.item(0));
            }

            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION,
                    "yes");
            transformer.transform(new DOMSource(fragment), new StreamResult(
                    result));
            return result.toString("UTF-8");
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "error!" + result.toString();
    }

    public String getDescription(String prefix, String postfix, String basePath) {
        return getDescription(
                new File(
                        ((com.unitesk.requality.nodetypes.Document) getParent())
                                .getHTMLPath()).getParent(),
                super.getDescription(), prefix, postfix, basePath);
    }

    private static String calcRelPath(String htmlPath, String basePath) {
        try {
            return getRelativeFile(new File(htmlPath), new File(basePath))
                    .getPath();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    private static void walk(Node node, String relPath, String htmlFolderPath) {
        proccess(node, relPath, htmlFolderPath);
        for (int i = 0; i < node.getChildNodes().getLength(); i++) {
            walk(node.getChildNodes().item(i), relPath, htmlFolderPath);
        }
    }

    private static void proccess(Node node, String relPath,
            String htmlFolderPath) {
        String nodeName = node.getNodeName().toLowerCase();
        if (nodeName.equals("img")) {
            Node attr = node.getAttributes().getNamedItem("src");
            if (attr != null) {
                if (new File(htmlFolderPath + File.separator
                        + attr.getNodeValue()).exists()) {
                    String newPath = relPath + attr.getNodeValue();
                    attr.setNodeValue(newPath);
                }
            }
        }
    }

    private static String toString(TreeDB db, String locQId) {
        Requirement reqs[] = db.getRequirementsFor(locQId);
        if (reqs.length == 0) {
            return "orphan location(" + locQId + ")";
        }
        String[] names = new String[reqs.length];
        Document doc = (Document) db.getNode(db.getParentQualifiedId(locQId));

        int i = 0;
        List<UUID> locs = doc.getLocations();

        for (Requirement curReq : reqs) {
            String reqName = curReq.getUserFriendlyName();
            int count = 0;
            int cur = 0;
            for (UUID uuid : locs) {
                String lqid = doc.getQualifiedId() + "/" + uuid.toString();
                Requirement[] lreqs = db.getRequirementsFor(lqid);
                if (lreqs != null && Arrays.asList(lreqs).contains(curReq)) {
                    count++;
                }
                if (lqid.equals(locQId)) {
                    cur = count;
                }
            }

            // TODO
            // Assert.isTrue(count != 0);
            if (count == 1) {
                names[i] = reqName;
            } else {
                names[i] = reqName + "#" + cur;
            }
            i++;
        }

        StringBuilder sb = new StringBuilder();
        sb.append(names[0]);
        for (i = 1; i < names.length; i++) {
            sb.append(", ");
            sb.append(names[i]);
        }
        return sb.toString();
    }

    @Override
    public String toString() {
    	if(label==null) {
    		label = Location.toString(getTreeDB(), getQualifiedId());
    	}
        return label;
    }

    /**
     * Returns the path of one File relative to another.
     * 
     * @param target
     *            the target directory
     * @param base
     *            the base directory
     * @return target's path relative to the base directory
     * @throws IOException
     *             if an error occurs while resolving the files' canonical names
     */
    public static File getRelativeFile(File target, File base)
            throws IOException {
        String[] baseComponents = base.getCanonicalPath().split(
                Pattern.quote(File.separator));
        String[] targetComponents = target.getCanonicalPath().split(
                Pattern.quote(File.separator));
        // skip common components
        int index = 0;
        for (; index < targetComponents.length && index < baseComponents.length; ++index) {
            if (!targetComponents[index].equals(baseComponents[index]))
                break;
        }
        StringBuilder result = new StringBuilder();
        if (index != baseComponents.length) {
            // backtrack to base directory
            for (int i = index; i < baseComponents.length; ++i)
                result.append(".." + File.separator);
        }
        for (; index < targetComponents.length; ++index)
            result.append(targetComponents[index] + File.separator);
        if (!target.getPath().endsWith("/") && !target.getPath().endsWith("\\")) {
            // remove final path separator
            result.delete(result.length() - "/".length(), result.length());
        }
        return new File(result.toString());
    }

    protected static final String ATTR_HIDDEN = "_hidden";

    public boolean isHidden() {
        return getBooleanAttributeValue( ATTR_HIDDEN );
    }

    public void isHidden(boolean hidden) {
        putAttribute(new Attribute(AttributeType.BOOL, ATTR_HIDDEN,hidden));
    }

	public void resetLabel() {
		label = null;
	}
}
